#include "effects.h"

#include <num/num.h>
#include "../utilities/fast_trig.h"

#include <cstring>

/********************************\
 *   MOPRHING EFFECT - MODULE 2   *
 * Specific data and structures *
\********************************/

typedef struct
{
  libnum::num32 x;
  libnum::num32 y;
  libnum::num32 z;
  libnum::num32 rx;
  libnum::num32 ry;
  libnum::num32 rz;
  int16_t sx;
  int16_t sy;
} Vertex3D;

#define NB_POINTS 2048
#define SCALE 80

Vertex3D Sphere[NB_POINTS];
Vertex3D Torus[NB_POINTS];
Vertex3D Cylinder[NB_POINTS];
Vertex3D Morph[NB_POINTS];

libnum::num32 scale;
libnum::num32 eye;
libnum::num32 eyescale;
int centerx, centery;

static float anglef = 0.0;
static int angle = 0;

static double frame_counter = 0;


void ProjectVertex( Vertex3D *vertex )
{
	libnum::num32 divisor = libnum::num32(1) / (scale * vertex->rz + eye);

    vertex->sx = centerx + (int) ((eyescale * vertex->rx) * divisor );
	vertex->sy = centery + (int) ((eyescale * vertex->ry) * divisor );
}


void RotateVertex( Vertex3D *vertex, int ax, int ay, int az )
{
	// Rotate around x axis
	vertex->ry = vertex->y * FastCosInt(ax) - vertex->z * FastSinInt(ax);
	vertex->rz = vertex->y * FastSinInt(ax) + vertex->z * FastCosInt(ax);

	// Rotate around y axis
	vertex->rx = vertex->x * FastCosInt(ay) + vertex->rz * FastSinInt(ay);
	vertex->rz = vertex->x * -FastSinInt(ay) + vertex->rz * FastCosInt(ay);

	// Rotate around z axis
	libnum::num32 tmpx = vertex->rx * FastCosInt(az) - vertex->ry * FastSinInt(az);
	vertex->ry = vertex->rx * FastSinInt(az) + vertex->ry * FastCosInt(az);
	vertex->rx = tmpx;
}


void morph_init( bopti_image_t *screen )
{
    eye = libnum::num32( 250 );
    scale = libnum::num32( SCALE );
    eyescale = eye*scale;

    centerx = (DWIDTH / 2);
    centery = (MAXHEIGHT / 2);

    // Create Sphere
	for (int i = 0; i < NB_POINTS; i++)
    {
		int theta = rand() % 360;
		int phi = rand() % 360 ;
		Sphere[i].x = FastCosInt(phi) * FastSinInt(theta);
		Sphere[i].y = FastSinInt(phi) * FastSinInt(theta);
		Sphere[i].z = FastCosInt(theta);
	}

	// Create torus
	for (int i = 0; i < NB_POINTS; i++)
    {
        int theta = rand() % 360;
		int phi = rand() % 360 ;
		Torus[i].x = (libnum::num32(0.8) + libnum::num32(0.4) * FastCosInt(theta)) * FastCosInt(phi);
		Torus[i].y = (libnum::num32(0.8) + libnum::num32(0.4) * FastCosInt(theta)) * FastSinInt(phi);
		Torus[i].z = libnum::num32(0.4) * FastSinInt(theta);
	}

	// Create cylinder
	for (int i = 0; i < NB_POINTS; i++)
    {
        int theta = rand() % 360;
		int height = rand() % 800;
		Cylinder[i].x = FastCosInt(theta);
		Cylinder[i].y = FastSinInt(theta);
		Cylinder[i].z = libnum::num32( ((float) height - 400.0f) / 500.0f );
	}
}


void morph_update( bopti_image_t *screen, [[Maybe_unused]] float dt )
{
    anglef += dt / 10000.0;
    angle = (int) anglef;

    frame_counter += dt / 10000.0;
    int frame = (int)frame_counter % 1536;

	// Transform the torus into a sphere
	if (frame >= 0 && frame < 256) {
		int key = frame;
		for (int i = 0; i < NB_POINTS; i++) {
			Morph[i].x = (libnum::num32(key) * Sphere[i].x + libnum::num32(256 - key) * Torus[i].x) / libnum::num32(256);
			Morph[i].y = (libnum::num32(key) * Sphere[i].y + libnum::num32(256 - key) * Torus[i].y) / libnum::num32(256);
			Morph[i].z = (libnum::num32(key) * Sphere[i].z + libnum::num32(256 - key) * Torus[i].z) / libnum::num32(256);
		}
	}

	// Transform the sphere into a torus
	if (frame >= 512 && frame <768 ) {
		int key = frame - 512;
		for (int i = 0; i < NB_POINTS; i++) {
			Morph[i].x = (libnum::num32(key) * Cylinder[i].x + libnum::num32(256 - key) * Sphere[i].x) / libnum::num32(256);
			Morph[i].y = (libnum::num32(key) * Cylinder[i].y + libnum::num32(256 - key) * Sphere[i].y) / libnum::num32(256);
			Morph[i].z = (libnum::num32(key) * Cylinder[i].z + libnum::num32(256 - key) * Sphere[i].z) / libnum::num32(256);
		}
	}

	// Transform the sphere into a torus
	if (frame >= 1024 && frame <1280 ) {
		int key = frame - 1024;
		for (int i = 0; i < NB_POINTS; i++) {
			Morph[i].x = (libnum::num32(key) * Sphere[i].x + libnum::num32(256 - key) * Cylinder[i].x) / libnum::num32(256);
			Morph[i].y = (libnum::num32(key) * Sphere[i].y + libnum::num32(256 - key) * Cylinder[i].y) / libnum::num32(256);
			Morph[i].z = (libnum::num32(key) * Sphere[i].z + libnum::num32(256 - key) * Cylinder[i].z) / libnum::num32(256);
		}
	}

	// Rotate the vertex in world coordinates and then project them on screen
    for (int i = 0; i < NB_POINTS; i++)
    {
		RotateVertex(&Morph[i], 0, angle, angle);
		ProjectVertex(&Morph[i]);
	}
}


void morph_render( bopti_image_t *screen )
{
	for (int i = 0; i < NB_POINTS; i++)
		dpixel( Morph[i].sx, Morph[i].sy, C_WHITE);
}

void morph_deinit( bopti_image_t *screen )
{

}