#include <gint/display.h>
#include <gint/keyboard.h>
#include <gint/dma.h>
#include <gint/rtc.h>
#include <gint/clock.h>

#include <fxlibc/printf.h>

#include "config.h"

#include <libprof.h>
#include <gint/kmalloc.h>

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "utilities/fast_trig.h"
#include "utilities/extrakeyboard.h"
#include "utilities/utilities.h"


#include "effects/effects.h"




#define TOTAL_NB_EFFECT 4
typedef void (*PtrToFunc)( bopti_image_t * );
typedef void (*PtrToFuncDt)( bopti_image_t *, float );

PtrToFunc Effects_Init[] = { &plasma_init, &morph_init, &matrix_init, &lens_init };
PtrToFuncDt Effects_Update[] = { &plasma_update, &morph_update, &matrix_update, &lens_update };
PtrToFunc Effects_Render[] = { &plasma_render, &morph_render, &matrix_render, &lens_render };
PtrToFunc Effects_DeInit[] = { &plasma_deinit, &morph_deinit, &matrix_deinit, &lens_deinit };
int current_effect = 0;


#if(MORE_RAM)
    static kmalloc_arena_t extended_ram = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, 0, 0,0,0,0,0,0 };
    static kmalloc_arena_t *_uram;
    kmalloc_gint_stats_t *_uram_stats;
    kmalloc_gint_stats_t *extram_stats;
#else
    static kmalloc_arena_t *_uram;
    kmalloc_gint_stats_t *_uram_stats;
#endif



bool exitToOS = false;
bool skip_intro = false;
bool screenshot = false;
bool record = false;

bool choosePrevious = false;
bool chooseNext = false;
int reverse = false;
float timeSinceLastEffect = 0.0f;

KeyboardExtra Keyboard;


float elapsedTime = 0.0f;
uint32_t time_update=0, time_render=0;
prof_t perf_update, perf_render;


extern font_t font_matrix;



bopti_image_t *screen;
uint16_t palette[256];
size_t palette_size = 512;
uint8_t imagedata[DWIDTH*MAXHEIGHT];



int get_pixel(bopti_image_t const *img, int x, int y)
{
    return image_get_pixel(img, x, y);
}

void set_pixel(bopti_image_t *img, int x, int y, int color)
{
    image_set_pixel(img, x, y, color);
}


bool AddMoreRAM( void )
{
    #if(MORE_RAM && !CALCEMU) 

        /* allow more RAM */
        char const *osv = (char*) 0x80020020;

        if((!strncmp(osv, "03.", 3) && osv[3] <= '6') && gint[HWCALC] == HWCALC_FXCG50) // CG-50
        {
            extended_ram.name = "extram";
            extended_ram.is_default = true;
            extended_ram.start =   (void *)0x8c200000;
            extended_ram.end = (void *)0x8c4e0000 ;

            kmalloc_init_arena(&extended_ram, true);
            kmalloc_add_arena(&extended_ram );
            return true;
        }
        else if (gint[HWCALC] == HWCALC_PRIZM)  // CG-10/20
        {

            extended_ram.name = "extram";
            extended_ram.is_default = true;

            uint16_t *vram1, *vram2;
            dgetvram(&vram1, &vram2);
            dsetvram(vram1, vram1);

            extended_ram.start = vram2;
            extended_ram.end = (char *)vram2 + 396*224*2;

            kmalloc_init_arena(&extended_ram, true);
            kmalloc_add_arena(&extended_ram );
            return false;

        }
        else if (gint[HWCALC] == HWCALC_FXCG_MANAGER) // CG-50 EMULATOR
        {

            extended_ram.name = "extram";
            extended_ram.is_default = true;
            extended_ram.start =   (void *)0x88200000;
            extended_ram.end = (void *)0x884e0000 ;

            kmalloc_init_arena(&extended_ram, true);
            kmalloc_add_arena(&extended_ram );
            return true;
        }
    #elif (MORE_RAM && CALCEMU)

        extended_ram.name = "extram";
        extended_ram.is_default = true;
        extended_ram.start =   (void *)0x8c200000;
        extended_ram.end = (void *)0x8c4e0000 ;

        kmalloc_init_arena(&extended_ram, true);
        kmalloc_add_arena(&extended_ram );
        return true;

    #else

        return false;

    #endif

    return false;
}

void FreeMoreRAM( void )
{
    #if(MORE_RAM)
        memset(extended_ram.start, 0, (char *)extended_ram.end - (char *)extended_ram.start);
    #endif
}



static void Initialise( void )
{
    (*Effects_Init[ current_effect ])( screen );
}

static void Render( void )
{
    (*Effects_Render[ current_effect ])( screen );
}

static void Update( [[maybe_unused]] float dt )
{
    (*Effects_Update[ current_effect ])( screen, dt );
}

static void Close( void )
{
    (*Effects_DeInit[ current_effect ])( screen );
}



static void GetInputs( [[maybe_unused]] float dt  )
{
    if (Keyboard.IsKeyPressed(MYKEY_EXIT)) {exitToOS = true; };

    if ((Keyboard.IsKeyPressedEvent(MYKEY_F1) || choosePrevious) && current_effect>=1)
    {
        if (choosePrevious) choosePrevious = false;
        
        timeSinceLastEffect = 0.0f;
        
        // close the current effect by calling its DeInitFunction
        Close();
        // switch to new effect
        current_effect--;
        // initialise the current effect
        Initialise();
    }

    if ((Keyboard.IsKeyPressedEvent(MYKEY_F6) || chooseNext) && current_effect<=TOTAL_NB_EFFECT-2)
    {
        if (chooseNext) chooseNext = false;

        timeSinceLastEffect = 0.0f;

        // close the current effect by calling its DeInitFunction
        Close();
        // switch to new effect
        current_effect++;
        // initialise the current effect
        Initialise();
    }
}



int main(void)
{

    _uram = kmalloc_get_arena("_uram");
    #if(MORE_RAM)
        bool canWeAllocateMoreRam = AddMoreRAM();
    #endif
    

    int EntryClockLevel;
    EntryClockLevel = clock_get_speed();
    clock_set_speed( CLOCK_SPEED_F4 );

    srand( rtc_ticks() );

    screen = image_alloc( DWIDTH, MAXHEIGHT, IMAGE_P8_RGB565 );
    screen->width = DWIDTH;
    screen->height = MAXHEIGHT;

    exitToOS = false;

    Initialise();

    __printf_enable_fp();
    __printf_enable_fixed();

    prof_init();

    do
	{
        perf_update = prof_make();
        prof_enter(perf_update);
		{
			// all the stuff to be update should be put here
            Keyboard.Update( elapsedTime );

			// read inputs from the player
			GetInputs( elapsedTime );

			// update as per the time spend to do the loop
			Update( elapsedTime );

            // update the RAM consumption status
            _uram_stats = kmalloc_get_gint_stats(_uram);
            #if(MORE_RAM)
                extram_stats = kmalloc_get_gint_stats(&extended_ram);
            #endif
		}
        prof_leave(perf_update);
        time_update = prof_time(perf_update);

        perf_render = prof_make();
        prof_enter(perf_render);
		{
			// all the stuff to be rendered should be put here
			dclear( C_BLACK );
            
            Render();
   			
            dupdate();
		}
        prof_leave(perf_render);
        time_render = prof_time(perf_render);
		
        elapsedTime = ((float) (time_update+time_render));

        timeSinceLastEffect += elapsedTime / 1000000.0f;
        if (timeSinceLastEffect >= 15.0f)
        {
            if(!reverse)
            {
                if(current_effect<TOTAL_NB_EFFECT-1) chooseNext=true;
                else
                {
                    reverse = true;
                    choosePrevious = true;
                }
            }
            else
            {
                if(current_effect>0) choosePrevious=true;
                else
                {
                    reverse = false;
                    chooseNext = true;
                }
            }
        }


    }
    while (exitToOS==false);


    prof_quit();

    clock_set_speed( EntryClockLevel );

    image_free( screen );

    #if(MORE_RAM)
        if (canWeAllocateMoreRam) FreeMoreRAM( );
    #endif

    return 1;
}
